/***********************************************************************************************************************
* @file beachhead.h
*
* @brief This file provides API interface definition of beachhead blockchain engine.
*
* (C) Copyright 2021 SmartAxiom,Inc All Rights Reserved
*
*  SMARTAXIOM CONFIDENTIAL
*  The source code contained or described herein and all documents related to
*  the source code ("Material") are owned by SmartAxiom,Inc or its
*  suppliers or licensor's. Title to the Material remains with SmartAxiom, Inc
*  or its suppliers and licensor's. The Material contain trade secrets and proprietary
*  and confidential information of SmartAxiom or its suppliers and licensor's.
*  The Material is protected by worldwide copyright and trade secret laws and
*  treaty provisions. No part of the Material may be used, copied, reproduced, modified,
*  published, uploaded, posted,transmitted, distributed, or disclosed in any way without
*  SmartAxiom prior express written permission.
*
*  No license under any patent, copyright, trade secret or other intellectual
*  property right is granted to or conferred upon you by disclosure or delivery
*  of the Materials, either expressly, by implication, inducement, estoppel or
*  otherwise. Any license under such intellectual property rights must be
*  express and approved by SmartAxiom in writing.
*
*/


#ifndef BEACHHEAD_H_
#define BEACHHEAD_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <time.h>
#include <stdlib.h>
#include "error.h"

#define SYNC_ENVELOPE_VERSION     2       /* current protocol version*/
#define SYNC_ENVELOPE_PLAIN       0x01    /* plain protocol without signatures*/
#define SYNC_ENVELOPE_SIGNED      0x02    /* signed messages (unchained) */
#define SYNC_ENVELOPE_CHAINED     0x03    /* chained signed messages */
#define SYNC_ENVELOPE_BINAR       0x00    /* Binary number or unknown values */

/**
 *  Frame type : BINAR, SIGNED, PLAIN and CHAINED
 */
typedef enum {
    BINAR      = ((SYNC_ENVELOPE_VERSION << 4) | SYNC_ENVELOPE_BINAR),   		/*Binary block*/
    SIGNED     = ((SYNC_ENVELOPE_VERSION << 4) | SYNC_ENVELOPE_SIGNED),  		/*Block Signed and encrypted*/
    PLAIN_TXT  = ((SYNC_ENVELOPE_VERSION << 4) | SYNC_ENVELOPE_PLAIN),   		/*Block Plain txt*/
    SYNC_ENVELOPE_CODE = ((SYNC_ENVELOPE_VERSION << 4) | SYNC_ENVELOPE_CHAINED) /*Block signed and chain code*/
}sync_envelope_frame_type;

/**
 * sync_envelope carrying block which depend upon message type,
 * maximum message type is 0 to 255
 */

typedef enum{
    DEVICE_DISCOVERY=0,              /* Session Discovery */
    DEVICE_HANDSHAKE=1,              /* Session Create */
    DEVICE_SESSION_VERIFY=2,         /* Session Verify */
	DEVICE_NEW=3,                    /* New Identity */
	DEVICE_REGISTER=4,               /* Identity Register */
	DEVICE_ACTIVATE=5,               /* Identity Activate  */
    DEVICE_SEND_DATA=6,              /* Generate Sync Envelope Block */
    DEVICE_AUTHENTICATE=7,           /* Authenticate Identity */
    DEVICE_SYNC=8,                   /* Sync Identity last Block*/
	DEVICE_UNAUTHENTICATE=9,         /* Unauthenticated Identity*/
    DEVICE_UPGRADE=10,               /* Upgrade firmware */
    DEVICE_RECEVIED=11,              /* Verify Received Block */
	DEVICE_HEARTBEAT=12,             /* Heart Beat block */
	DEVICE_STATUS=13,                /* Device Status */
	DEVICE_REQ_ACTIVATION=14         /* request for device activation*/
} sync_envelope_msg_type;


#define SYNC_ENVELOPE_SZ                   512    /* Sync_Envelope block size with frame_type chain */
#define SYNC_ENVELOPE_KEY_SZ               32     /* public key size*/
#define SYNC_ENVELOPE_SIGN_SZ              64     /* signatures has 64 bytes (same as size of hash) */
#define SYNC_ENVELOPE_AESKEY_SZ            16     /* AES 128 Key size */
#define SYNC_ENVELOPE_SIZEOFIV_HEX         16     /* IV Size */
#define SYNC_ENVELOPE_PAYLOAD_SZ           256    /* Payload Size 257 byte*/
#define SYNC_ENVELOPE_CRC_SZ               4      /* CRC size 4 byte */
#define SYNC_ENVELOPE_SENSOR_SZ            16     /* Sensor name as string in payload */
#define SYNC_ENVELOPE_TS_SZ                10     /* Time stamp size  */
#define SYNC_ENVELOPE_UUID                 32     /* Device UUID size */
#define SYNC_ENVELOPE_HASH_SZ              64     /* SHH 256 */


#define PROVISION_MSG "48656c6c6f20426c6f636b4c6f636b0d0a"

#define SYNC_ENVELOPE_ENCRYPT_SZ sizeof(sync_envelope_frame_type) \
                         + sizeof(sync_envelope_msg_type) \
                         + SYNC_ENVELOPE_TS_SZ \
                         + SYNC_ENVELOPE_UUID \
                         + SYNC_ENVELOPE_SIGN_SZ \
                         + sizeof(uint16_t) \
						 + BH_SZ_IV_HEX \
                         + SYNC_ENVELOPE_SIGN_SZ \
						 + SYNC_ENVELOPE_CRC_SZ

/* Sync_Envelope block size with frame_type plain text */
#define SYNC_ENVELOPE_PLAIN_TXT_SZ sizeof(sync_envelope_frame_type) \
                         + sizeof(sync_envelope_msg_type) \
                         + SYNC_ENVELOPE_TS_SZ \
                         + SYNC_ENVELOPE_UUID \
                         + SYNC_ENVELOPE_SIGN_SZ \
                         + sizeof(uint16_t) \
					     + SYNC_ENVELOPE_SIGN_SZ \
						 + SYNC_ENVELOPE_CRC_SZ

/**
 * Initialize rtc controller
 * @return
 */
typedef int (*rtc_init_ctrl)(void);

/**
 * set RTC clock with time_t
 * */
typedef void (*rtc_set_time) (time_t);

/**
 * Get RTC clock time
 */
typedef time_t (*rtc_get_time) (void);


/**
 * non-volatile storage read and write
 */
typedef int (*nvmem_read)(uint8_t *pvDest, uint32_t pvOffset, uint32_t ulByte);
typedef int (*nvmem_write)(uint8_t *pvSrc, uint32_t pvOffset, uint32_t ulByte);
typedef int (*nvmem_init)(void);


/**
 *  Sync envelope block which hold
 *  frame_type         : sync envelope block frame type
 *  message type       : sync envelope block message Type
 *  block time stamp   : sync envelope block time stamp
 *  previous signature : previous block signature
 *  payload_sz         : Max payload size
 *  alloc_sz           : allocate payload size, should be less then payload size
 *  payload            : payload context
 *  signature          : 512 bit, 64-byte array (ED25519 signature, 64 bytes)
 *  CRC                : 4 byte based crc
 */
typedef struct  sync_envelope_blk {
        uint8_t   frame_type ;
        uint8_t   msg_type;
        uint8_t   timestamp[SYNC_ENVELOPE_TS_SZ];
        uint8_t   device_UUID [SYNC_ENVELOPE_UUID];
        uint8_t   previous_signature[SYNC_ENVELOPE_SIGN_SZ];
        uint32_t  payload_len;
        uint8_t   *payload;
        uint8_t   signature[SYNC_ENVELOPE_SIGN_SZ];
        uint16_t  crc;
} __attribute__((packed)) sync_envelope;


/**
 * Manufacture information
 */
#define UNIQUE_CLIENT_ID_BYTES	 32
#define BYTES 					 16
#define MAC_BYTES_SZ 			  6
#define MFG_INFO_SZ UNIQUE_CLIENT_ID_BYTES \
                    + (MAC_BYTES_SZ*2 + 5) + BYTES + BYTES + BYTES\
					+ sizeof(uint32_t) + sizeof(uint16_t) + sizeof(uint16_t)

typedef struct mfg_info {
	   uint8_t  client_id[UNIQUE_CLIENT_ID_BYTES];
	   uint8_t  mac_addr[MAC_BYTES_SZ];
       uint8_t  mfg_name[BYTES];
       uint8_t  model_name[BYTES];
       uint8_t  serial_number[BYTES];
       uint32_t sw_ver;
       uint16_t boot_rom_ver;
       uint16_t Firmware;
 }__attribute__((packed))mfg_info_t;


/**
 * Device Manufacture information
 * */
typedef struct sync_evelope_en_blk {
	uint16_t       input_blk_len;
	uint8_t       *input_blk;
	uint16_t       out_buffer_len;
	uint8_t       *out_buffer;
	uint8_t        state;
}__attribute__((packed)) sync_envelope_en_blk_t;

/**
 * Initialize axiom protocol data that include RTC interface and NV RAM
 */
typedef struct axiom_protocol_data {
    uint8_t      device_UUID [SYNC_ENVELOPE_UUID];
    uint32_t     InMemory_ledger_sz;
    mfg_info_t   *pMfg_info;
    rtc_init_ctrl	rtc_init;
    rtc_get_time get_time;
    rtc_set_time set_time;
    nvmem_read   nv_read;
    nvmem_write  nv_write;
    nvmem_init   nv_init;
}__attribute__((packed)) axiom_protocol_data_t;


/**
 * Create new axiom protocol context. Allocates memory for context and data buffer on heap,
 * initializes and sets user sign callback and create device token and set previous token to 0.
 *
 * */
void *sync_envelope_init(const void *);

/**
 * The write payload function add a new payload in sync_envelope
 *  this function is called from #axiom_protocol_add_message
 *  @param bh_sync_context context of axiom protocol
 *  @param sync_env envelope block
 *  @param sync_envelope_msg_type payload msg type
 */
int sync_envelope_process_payload(void *bh_sync_context,uint8_t frame_type, uint8_t msg_type,
                                  void *payload, uint32_t payload_len, uint8_t *out_buffer);

/**
 * The verification function to check the validity of the message.
 * This function is called from #axiom_protocol_verify
 *
 * @param buf the data to verify
 * @param len the length of the data buffer
 * @param signature the signature to check the data with (64 byte)
 * @return 0 on success
 * @return -1 if the verification failed
 */

int sync_envelope_verified_buffer(const void *bh_context,sync_envelope_en_blk_t *synv_env_blk);


/**
 *  Free allocate buffer for sync envelope
 */
void sync_envelope_free(void *bh_context);

/**
 * @brief
 * @param bh_context
 * @param new_state
 * @return
 */

sync_envelope_err_type sync_envelope_set_state(const void *bh_context,
		                                       sync_envelope_msg_type new_state);

/**
 * @brief
 * @param  bh_context beachhead handler
 * @return current state of beachhead library
 */
sync_envelope_msg_type sync_envelope_get_state(const void *bh_context);




#ifdef __cplusplus
}
#endif


#endif /* BEACHHEAD_H_*/
