/***********************************************************************************************************************
 * File Name    : net_thread_entry.c
 * Description  : This file contains the publisher code for the beachhead-SDK

 * (C) Copyright 2021 SmartAxiom,Inc All Rights Reserved
 *
 *  SMARTAXIOM CONFIDENTIAL
 *  The source code contained or described herein and all documents related to
 *  the source code ("Material") are owned by SmartAxiom,Inc or its
 *  suppliers or licensor's. Title to the Material remains with SmartAxiom, Inc
 *  or its suppliers and licensor's. The Material contain trade secrets and proprietary
 *  and confidential information of SmartAxiom or its suppliers and licensor's.
 *  The Material is protected by worldwide copyright and trade secret laws and
 *  treaty provisions. No part of the Material may be used, copied, reproduced, modified,
 *  published, uploaded, posted,transmitted, distributed, or disclosed in any way without
 *  SmartAxiom prior express written permission.
 *
 *  No license under any patent, copyright, trade secret or other intellectual
 *  property right is granted to or conferred upon you by disclosure or delivery
 *  of the Materials, either expressly, by implication, inducement, estoppel or
 *  otherwise. Any license under such intellectual property rights must be
 *  express and approved by SmartAxiom in writing.
 *
 * *******************************************************************************************/
#include <net_thread.h>
#include "beachhead_config.h"
#include "beachhead-app.h"
#include "usr_app.h"

/*Global variable of User Applications*/
bool g_beachhead_common_initialized;
void *bh_context = NULL;
bh_socket_t bh_socket = { RESET_VALUE };
static uint8_t sensor_payload[128] = { RESET_VALUE };


/**
 * Manufacture information
 * */
const mfg_info_t mfg_info =
{
		.client_id = "Renesas-Semiconductor",
		.mac_addr = { 0x44, 0x55, 0x22, 0x33, 0x00, 0x11 },
		.mfg_name = "SmartAxiom,Inc",
		.model_name = "EK-RA6M3",
		.serial_number = "R7FA6M3AH32BD",
		.sw_ver = 2,
		.boot_rom_ver = 1,
		.Firmware = 1.00
};

/**
 * @brief Beachhhead-SDK initialization function which include
 *        RTC,flush read and write and unique ID
 * @return
 */
static int beachhead_init()
{
	int ret = SYNC_ERR_FAIL;
	bsp_unique_id_t *uuid = NULL;

	/* *
	 * Initialize application data include RTC,NV memory and
	 * Unique ID of the device
	 * */
	axiom_protocol_data_t axiom_data;

	/*Initialize RTC time function*/
	axiom_data.rtc_init = rtc_user_init;
	axiom_data.get_time = get_user_time;
	axiom_data.set_time = set_user_time;


	/*Initialized nvram read and write function*/
	axiom_data.nv_init = nvm_data_flash_init;
	axiom_data.nv_read = nvm_data_flash_read;
	axiom_data.nv_write = nvm_data_flash_write;

	/*Initialize device information*/
	axiom_data.pMfg_info = (mfg_info_t*) &mfg_info;

	/*Initialize Unique Device ID*/
	uuid = (bsp_unique_id_t *) R_BSP_UniqueIdGet();
	AsciitoHex(&axiom_data.device_UUID[0],uuid->unique_id_bytes,
			     sizeof(uuid->unique_id_bytes));
	APP_PRINT("\r\nUnique Device UUID:  %s", axiom_data.device_UUID);

	/*Initialize circular buffer*/
	axiom_data.InMemory_ledger_sz = 4;

	/*Initialize Beachhead SDK protocol*/
	bh_context = sync_envelope_init((void*) &axiom_data);

	if (bh_context == NULL) {
		APP_ERR_PRINT("\r\n Unable to Initialize Beachhead SDK....");
	} else {
		ret = SYNC_OK;
		APP_PRINT("\r\n Initialize Beachhead Blockchain Engine......")
	}

	return ret;
}

/**
 * @brief Application init function
 * @return
 */

static int app_init()
{
	BaseType_t status = pdFALSE;
	fsp_pack_version_t version = { RESET_VALUE };
	int ret = SYNC_ERR_FAIL;

	/* version get API for FLEX pack information */
	R_FSP_VersionGet(&version);

	/* Example Project information printed on the RTT */
	APP_PRINT(BANNER_INFO, BH_VERSION, version.major, version.minor,
			version.patch);

	/* Prints the Ethernet Configuration prior to the IP Init*/
	APP_PRINT(ETH_PREINIT);

	print_ipconfig();

	/* FreeRTOS IP Initialization: This init initializes the IP stack  */
	status = IP_Setup((uint8_t *)mfg_info.mac_addr);

	if (pdFALSE == status) {
		APP_ERR_PRINT("\r\nIP Setup Failed");
		APP_ERR_TRAP(status);
	} else {
		APP_PRINT(ETH_POSTINIT);

		/*Initialize Socket*/
		APP_PRINT("\r\n.........Opening Communication Channel.......\r\n");
		memset(&bh_socket, 0, sizeof(bh_socket_t));
		bh_socket.vPort = PORT_B;
		bh_socket.bBroadcast = 1;
		bh_socket.socket_type = FREERTOS_SOCK_DGRAM;
		bh_socket.socket_protocol = FREERTOS_IPPROTO_UDP;
		socket_init(&bh_socket);

		/*Initialize SDK-Beachhead */
		ret = beachhead_init();

		if (bh_context != NULL && ret == SYNC_OK) {

			/*Initialize ADC*/
			ret = temp_sensor_init();

			if (ret != FSP_SUCCESS)
				APP_ERR_PRINT("\r\nUnable to Initialize ADC err %d", ret);
		} else
			APP_PRINT("\r\nUnable to initialize Beachhead ......");


	}
	return ret;
}

/**
 * @brief ADC reading in certain interval and publish event to responders
 * @param xTimer
 */
void g_new_timer0_callback(TimerHandle_t xTimer)
{
	uint8_t bh_state;

	if (xTimer != NULL && g_beachhead_common_initialized) {

		/*read ADC value */
		read_ADC_Temp(sensor_payload);

		/*publish event to publish event queue*/
		bh_state = DEVICE_SEND_DATA;
		xQueueSend(g_publisher_queue, &bh_state, 0);
	}
}


/**
 * @brief Heart beat task used to implement two communication with device and responder
 * @param xTimer
 */
void g_hb_callback(TimerHandle_t xTimer) {
	 uint8_t bh_state;

	if (xTimer != NULL && g_beachhead_common_initialized) {

		/*publish heart beat event for listening the subscriber events*/
		bh_state = DEVICE_HEARTBEAT;
		xQueueSend(g_publisher_queue, &bh_state, 0);
	}
}

/*******************************************************************************************************************//**
 * @brief      This is the User Thread for the EP.
 * @param[in]  Thread specific parameters
 * @retval     None
 **********************************************************************************************************************/
void net_thread_entry(void *pvParameters)
{
	int ret = SYNC_ERR_FAIL;
	g_beachhead_common_initialized = false;
	uint8_t process_blk[SYNC_ENVELOPE_PAYLOAD_SZ] = { 0 };
	uint8_t bh_state = DEVICE_DISCOVERY;

	FSP_PARAMETER_NOT_USED(pvParameters);

	/*Initialize SDK-Beachhead and authenticate device or register as new device*/
	if (false == g_beachhead_common_initialized) {

		/* Initialize beachhead ADC application to demonstrate decentralized edge
		 * security
		 * */
		ret = app_init();

		/*stop all periodic timer*/
		xTimerStop(g_new_timer0, 0);

		if (ret == SYNC_OK) {

			/*Initialize beachhead-SDK init service */
			ret = beachhead_common_init_service(bh_context, &bh_socket, &bh_state);

			/*register new device*/
			switch(ret) {
			case  SYNC_BLK_REGISTERED:
				g_beachhead_common_initialized = true;
				break;
			case SYNC_BLK_AUTHENTICATE: {
				/*
				 * After authenticate device, it does established the secure channel with
				 * responders.
				 * */
				if (g_beachhead_common_initialized && bh_state == DEVICE_SEND_DATA) {
					xTimerStart(g_new_timer0, 0);
					APP_PRINT("\r\n Starting ADC...");
				} else {
					if (bh_state == DEVICE_SYNC) {
					   xQueueSend(g_publisher_queue, &bh_state, 0);
					}
				}
				break;
			 }
			 case SYNC_BLK_ACTIVATE:
				g_beachhead_common_initialized = true;
				APP_PRINT("\r\n Starting ADC...");
				if (bh_state == DEVICE_ACTIVATE) {
				   xTimerStart(g_new_timer0, 0);
				   bh_state  = DEVICE_SEND_DATA;
				}
			    break;
			 default :
				/*Unauthorized device*/
				APP_ERR_PRINT("\r\n Unable to verify trust device...");
				APP_ERR_TRAP(ret);
			}
		} else {
			APP_ERR_PRINT("\r\n Unable to Initialize Beachhead-SDK...");
			APP_ERR_TRAP(ret);
		}
	}

	/*enable subscriber threads*/

	if ((xTimerIsTimerActive(g_hb_timer1) == pdFALSE)
			&& g_beachhead_common_initialized) {
    				xTimerStart(g_hb_timer1,0);
	}


	/*starting publisher thread*/
	while (true) {
	   xQueueReceive(g_publisher_queue, &bh_state, portMAX_DELAY);
	   /*publish events*/
	   switch (bh_state) {
	   /*send data block*/
		case DEVICE_SEND_DATA: {
		   ret = beachhead_send_data(bh_context, sensor_payload,(uint16_t)
				   strlen((char*) sensor_payload), process_blk,&bh_socket, &bh_state);
		   if (ret == SYNC_OK) {
			   APP_PRINT("\r\nPublish Sync Block to responders...");
		   } else {
				/*Any Communication errors will fall here and reestablish communication*/
				APP_ERR_PRINT(	"\r\nUnable to communicate with responder Error: %d....",ret);
			    /*stop ADC service*/
				g_beachhead_common_initialized = false;
			    beachhead_close(&bh_socket);
				/*
				 * After receiving any error, it does reestablished a new secure channel
				 * with  new responders
				 * */
			    bh_state= DEVICE_DISCOVERY;
				beachhead_common_init_service(bh_context, &bh_socket,&bh_state);
			}
		}
		break;
			/*sync device with responder*/
		case DEVICE_SYNC: {
			ret = beachhead_device_sync(bh_context, &bh_socket, &bh_state);
			if (ret == SYNC_OK) {
				/*start temperature sensor reading task */
				g_beachhead_common_initialized = true;
				APP_PRINT("\r\nPublish Device sync request to responders...");
			} else {
				/*Any Communication errors will fall here and reestablish communication*/
				APP_ERR_PRINT("\r\nUnable to communicate with responder, Error: %d....",ret);
				/*stop ADC service*/
				//xTimerStop(g_new_timer0, 0);
				g_beachhead_common_initialized = false;
				beachhead_close(&bh_socket);
				/*
				 * After receiving any error, it does reestablished a new secure channel
				 * with  new responders
				 * */
				bh_state= DEVICE_DISCOVERY;
				beachhead_common_init_service(bh_context, &bh_socket,&bh_state);
			}
		}
	    break;
		case DEVICE_STATUS: {
		  APP_PRINT("\r\nPublish Device status...");
		}
		break;
		case DEVICE_ACTIVATE: {
			 APP_PRINT("\r\nStarting new device activation ...");

			 ret = beachhead_activate_device(bh_context, &bh_socket,&bh_state);

			 if (ret == SYNC_OK) {
				APP_PRINT("\r\nDevice activation is completed...");
			 } else {
				APP_ERR_PRINT("\r\nUnable to activate new device...");
			 }
		}
		break;
			/*Send Heart Beat message to responders*/
		case DEVICE_HEARTBEAT:
			beachhead_heart_beat(bh_context, &bh_socket, &bh_state);
			break;
		default:
			/*Reset event queue and close open socket*/
			beachhead_close(&bh_socket);

			/*set initialized to false, so it can again to broadcast discovery messages*/
			g_beachhead_common_initialized = false;

			/*Reestablish connection with responder in local group*/
			bh_state= DEVICE_DISCOVERY;
			beachhead_common_init_service(bh_context, &bh_socket, &bh_state);
			break;
		}

	}
}

