/***********************************************************************************************************************
* @file  nvm_data_flash.c
*
* @brief flash read and write source files
*
* (C) Copyright 2021 SmartAxiom,Inc All Rights Reserved
*
*  SMARTAXIOM CONFIDENTIAL
*  The source code contained or described herein and all documents related to
*  the source code ("Material") are owned by SmartAxiom,Inc or its
*  suppliers or licensor's. Title to the Material remains with SmartAxiom, Inc
*  or its suppliers and licensor's. The Material contain trade secrets and proprietary
*  and confidential information of SmartAxiom or its suppliers and licensor's.
*  The Material is protected by worldwide copyright and trade secret laws and
*  treaty provisions. No part of the Material may be used, copied, reproduced, modified,
*  published, uploaded, posted,transmitted, distributed, or disclosed in any way without
*  SmartAxiom prior express written permission.
*
*  No license under any patent, copyright, trade secret or other intellectual
*  property right is granted to or conferred upon you by disclosure or delivery
*  of the Materials, either expressly, by implication, inducement, estoppel or
*  otherwise. Any license under such intellectual property rights must be
*  express and approved by SmartAxiom in writing.
*
*/

#include "nvm_data_flash.h"

flash_hp_instance_ctrl_t g_data_flash_ctrl;

const flash_cfg_t g_data_flash_cfg = { .data_flash_bgo = false, .p_callback =
		NULL, .p_context = NULL,
#if defined(VECTOR_NUMBER_FCU_FRDYI)
    .irq                 = VECTOR_NUMBER_FCU_FRDYI,
#else
		.irq = FSP_INVALID_VECTOR,
#endif
#if defined(VECTOR_NUMBER_FCU_FIFERR)
    .err_irq             = VECTOR_NUMBER_FCU_FIFERR,
#else
		.err_irq = FSP_INVALID_VECTOR,
#endif
		.err_ipl = (9), .ipl = (9), };

int nvm_data_flash_init(void)
{
	fsp_err_t err = FSP_SUCCESS;     // Error status

	err = R_FLASH_HP_Open(&g_data_flash_ctrl, &g_data_flash_cfg);
	return err;
}

int nvm_data_flash_write(uint8_t *pvSrc, uint32_t pvOffset, uint32_t ulByte)
{
	fsp_err_t err = FSP_SUCCESS;     // Error status

	/* Force write on block boundary */
	if ( pvOffset % NVM_DATA_FLASH_BLOCK_SIZE != 0 ) {
		return FSP_ERR_INVALID_ADDRESS;
	}
	/* Force write data is multiple of block size */
	if ( ulByte % NVM_DATA_FLASH_BLOCK_SIZE != 0 ) {
		return FSP_ERR_INVALID_SIZE;
	}
	/* Only allow writes in available space */
	if ( (pvOffset+ulByte) > NVM_DATA_FLASH_BYTES ) {
		return FSP_ERR_INVALID_SIZE;
	}

	err = R_FLASH_HP_Erase(&g_data_flash_ctrl, NVM_DATA_FLASH_START+pvOffset, ulByte/NVM_DATA_FLASH_BLOCK_SIZE);
	if (err != FSP_SUCCESS) {
		return err;
	}

	err = R_FLASH_HP_Write(&g_data_flash_ctrl, (uint32_t) pvSrc, NVM_DATA_FLASH_START+pvOffset, ulByte);


	return err;
}

int nvm_data_flash_read(uint8_t *pvDest, uint32_t pvOffset, uint32_t ulByte)
{
	/* Only allow read from available space */
	if ( (pvOffset+ulByte) > NVM_DATA_FLASH_BYTES ) {
		return FSP_ERR_INVALID_SIZE;
	}

	memcpy(pvDest, (uint8_t *) NVM_DATA_FLASH_START+pvOffset, ulByte);
	return FSP_SUCCESS;
}

void nvm_data_flash_test(uint32_t blk_num)
{

	uint8_t write_blk[NVM_DATA_FLASH_BLOCK_SIZE] = {0};
	uint8_t read_blk[NVM_DATA_FLASH_BLOCK_SIZE]= {0};
	fsp_err_t err = FSP_SUCCESS;     // Error status

	err = nvm_data_flash_init();
	if ( err != FSP_SUCCESS) {
		/* Init error */
		while(1);
	}

	/* Save existing data */
	err = nvm_data_flash_read(read_blk, blk_num * NVM_DATA_FLASH_BLOCK_SIZE, NVM_DATA_FLASH_BLOCK_SIZE);
	if ( err != FSP_SUCCESS) {
		/* Read error */
		while(1);
	}

	/* Write block */
	memset(write_blk, 0xA5, NVM_DATA_FLASH_BLOCK_SIZE);
	err = nvm_data_flash_write(write_blk,blk_num * NVM_DATA_FLASH_BLOCK_SIZE,NVM_DATA_FLASH_BLOCK_SIZE);
	if ( err != FSP_SUCCESS) {
		/* Write error */
		while(1);
	}

	/* Verify written data */
	memset(write_blk, 0x00, NVM_DATA_FLASH_BLOCK_SIZE);
	err = nvm_data_flash_read(write_blk, blk_num * NVM_DATA_FLASH_BLOCK_SIZE, NVM_DATA_FLASH_BLOCK_SIZE);
	if ( err != FSP_SUCCESS) {
		/* Read error */
		while(1);
	}
	for (int i=0; i<NVM_DATA_FLASH_BLOCK_SIZE; i++) {
		if ( write_blk[i] != 0xA5 ) {
			/* Verification failed */
			while(1);
		}
	}

	/* Restore block data */
	err = nvm_data_flash_write(read_blk,blk_num * NVM_DATA_FLASH_BLOCK_SIZE,NVM_DATA_FLASH_BLOCK_SIZE);
	if ( err != FSP_SUCCESS) {
		/* Write error */
		while(1);
	}
}
