/***********************************************************************************************************************
* @file  source file beachhead-api.c
*
* @brief  beachhead-SDK API interface
*
* (C) Copyright 2021 SmartAxiom,Inc All Rights Reserved
*
*  SMARTAXIOM CONFIDENTIAL
*  The source code contained or described herein and all documents related to
*  the source code ("Material") are owned by SmartAxiom,Inc or its
*  suppliers or licensor's. Title to the Material remains with SmartAxiom, Inc
*  or its suppliers and licensor's. The Material contain trade secrets and proprietary
*  and confidential information of SmartAxiom or its suppliers and licensor's.
*  The Material is protected by worldwide copyright and trade secret laws and
*  treaty provisions. No part of the Material may be used, copied, reproduced, modified,
*  published, uploaded, posted,transmitted, distributed, or disclosed in any way without
*  SmartAxiom prior express written permission.
*
*  No license under any patent, copyright, trade secret or other intellectual
*  property right is granted to or conferred upon you by disclosure or delivery
*  of the Materials, either expressly, by implication, inducement, estoppel or
*  otherwise. Any license under such intellectual property rights must be
*  express and approved by SmartAxiom in writing.
*
*/

#include "beachhead-api.h"
#include "common_utils.h"

/**
 *
 * @param bh_context
 * @param en_blk
 * @return
 */

int RetrieveBuffer(const void *bh_context, sync_envelope_en_blk_t *en_blk)
{
	int status;

	if (bh_context == NULL || en_blk == NULL)
		return SYNC_ERR_FAIL;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*Verified encrypt block */
	status = sync_envelope_verified_buffer(bh_context, en_blk);

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

/**
 *
 * @param bh_context
 * @param payload
 * @param payload_len
 * @param outBuffer
 * @param oBufferLen
 * @return
 */

sync_envelope_err_type SessionDiscovery(void *bh_context, uint8_t *payload, uint32_t payload_len,
		                                uint8_t *outBuffer, uint32_t *oBufferLen)
{
	sync_envelope_err_type status = SYNC_ERR_FAIL;
	uint32_t buf_len = 0;
	uint8_t frame_type = SYNC_ENVELOPE_SIGNED;
	uint8_t msg_type = DEVICE_DISCOVERY;

	if (bh_context == NULL || payload == NULL || outBuffer == NULL)
		return SYNC_ERR_INVALID_HANDLE;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*create discovery block*/
	buf_len = (uint32_t) sync_envelope_process_payload(bh_context, frame_type,
			msg_type, payload, payload_len, outBuffer);

	if (buf_len > 0) {
		status = SYNC_OK;
		*oBufferLen = buf_len;
	}

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

/**
 * @brief
 * @param bh_context
 * @param outBuffer
 * @param oBufferLen
 * @return
 */
sync_envelope_err_type CreateSession(void *bh_context, uint8_t *outBuffer, uint32_t *oBufferLen)
{
	sync_envelope_err_type status = SYNC_ERR_FAIL;
	uint32_t buf_len = 0;
	uint8_t frame_type = SYNC_ENVELOPE_SIGNED;
	uint8_t msg_type = DEVICE_HANDSHAKE;

	if (bh_context == NULL || outBuffer == NULL)
		return SYNC_ERR_INVALID_HANDLE;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*create discovery block*/
	buf_len = (uint32_t) sync_envelope_process_payload(bh_context, frame_type,msg_type, NULL, 0, outBuffer);

	if (buf_len > 0) {
		status = SYNC_OK;
		*oBufferLen = buf_len;
	}

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;

}

/**
 *
 * @param bh_context
 * @param outBuffer
 * @param oBufferLen
 * @return
 */

sync_envelope_err_type VerifySession(void *bh_context, uint8_t *outBuffer, uint32_t *oBufferLen)
{
	sync_envelope_err_type status = SYNC_ERR_FAIL;
	uint32_t buf_len = 0;
	uint8_t frame_type = SYNC_ENVELOPE_SIGNED;
	uint8_t msg_type = DEVICE_SESSION_VERIFY;

	if (bh_context == NULL || outBuffer == NULL)
		return SYNC_ERR_INVALID_HANDLE;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*create discovery block*/
	buf_len = (uint32_t) sync_envelope_process_payload(bh_context, frame_type, msg_type, NULL, 0, outBuffer);

	if (buf_len > 0) {
		status = SYNC_OK;
		*oBufferLen = buf_len;
	}

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

/**
 *
 * @param bh_context
 * @param outBuffer
 * @param oBufferLen
 * @return
 */

sync_envelope_err_type AuthenticateSession(void *bh_context, uint8_t *outBuffer, uint32_t *oBufferLen)
{
	sync_envelope_err_type status = SYNC_ERR_FAIL;
	uint32_t buf_len = 0;
	uint8_t frame_type = SYNC_ENVELOPE_SIGNED;
	uint8_t msg_type = DEVICE_AUTHENTICATE;

	if (bh_context == NULL || outBuffer == NULL)
		return SYNC_ERR_INVALID_HANDLE;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*create Authenticate block*/
	buf_len = (uint32_t) sync_envelope_process_payload(bh_context, frame_type, msg_type, NULL, 0, outBuffer);

	if (buf_len > 0) {
		status = SYNC_OK;
		*oBufferLen = buf_len;
	}

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;

}

/**
 *
 * @param bh_context
 * @param outBuffer
 * @param oBufferLen
 * @return
 */

sync_envelope_err_type AddNewIdentity(void *bh_context, uint8_t *outBuffer, uint32_t *oBufferLen)
{
	sync_envelope_err_type status = SYNC_ERR_FAIL;
	uint32_t buf_len = 0;
	uint8_t frame_type = SYNC_ENVELOPE_SIGNED;
	uint8_t msg_type = DEVICE_REGISTER;

	if (bh_context == NULL || outBuffer == NULL)
		return SYNC_ERR_INVALID_HANDLE;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*create Authenticate block*/
	buf_len = (uint32_t) sync_envelope_process_payload(bh_context, frame_type, msg_type, NULL, 0, outBuffer);

	if (buf_len > 0) {
		status = SYNC_OK;
		*oBufferLen = buf_len;
	}

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

/**
 *
 * @param bh_context
 * @param outBuffer
 * @param oBufferLen
 * @return
 */
sync_envelope_err_type ActivateIdentity(void *bh_context, uint8_t *outBuffer,
		uint32_t *oBufferLen) {
	sync_envelope_err_type status = SYNC_ERR_FAIL;
	uint32_t buf_len = 0;
	uint8_t frame_type = SYNC_ENVELOPE_SIGNED;
	uint8_t msg_type = DEVICE_ACTIVATE;

	if (bh_context == NULL || outBuffer == NULL)
		return SYNC_ERR_INVALID_HANDLE;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*create Authenticate block*/
	buf_len = (uint32_t) sync_envelope_process_payload(bh_context, frame_type, msg_type, NULL, 0, outBuffer);

	if (buf_len > 0) {
		status = SYNC_OK;
		*oBufferLen = buf_len;
	}

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

/**
 *
 * @param bh_context
 * @param payload
 * @param payload_len
 * @param outBuffer
 * @param oBufferLen
 * @return
 */

sync_envelope_err_type CreateSyncBlock(void *bh_context, uint8_t *payload,
		uint32_t payload_len, uint8_t *outBuffer, uint32_t *oBufferLen)
{
	sync_envelope_err_type status = SYNC_ERR_FAIL;
	uint32_t buf_len = 0;
	uint8_t frame_type = SYNC_ENVELOPE_SIGNED;
	uint8_t msg_type = DEVICE_SEND_DATA;

	if (bh_context == NULL || payload == NULL || outBuffer == NULL)
		return SYNC_ERR_INVALID_HANDLE;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*create discovery block*/
	buf_len = (uint32_t) sync_envelope_process_payload(bh_context, frame_type,
                                               			msg_type, payload, payload_len, outBuffer);

	if (buf_len > 0) {
		status = SYNC_OK;
		*oBufferLen = buf_len;
	}

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

/**
 *
 * @param bh_context
 * @param outBuffer
 * @param oBufferLen
 * @return
 */

sync_envelope_err_type ResyncSyncBlock(void *bh_context, uint8_t *outBuffer,
		uint32_t *oBufferLen) {
	sync_envelope_err_type status = SYNC_ERR_FAIL;
	uint32_t buf_len = 0;
	uint8_t frame_type = SYNC_ENVELOPE_SIGNED;
	uint8_t msg_type = DEVICE_SYNC;

	if (bh_context == NULL || outBuffer == NULL)
		return SYNC_ERR_INVALID_HANDLE;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*create Authenticate block*/
	buf_len = (uint32_t) sync_envelope_process_payload(bh_context, frame_type,msg_type, NULL, 0, outBuffer);

	if (buf_len > 0) {
		status = SYNC_OK;
		*oBufferLen = buf_len;
	}

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

/**
 *
 * @param bh_context
 * @return
 */

sync_envelope_err_type ReleaseContext(void *bh_context) {
	sync_envelope_err_type status = SYNC_ERR_FAIL;

	if (bh_context == NULL)
		return status;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	sync_envelope_free(bh_context);

	status = SYNC_OK;

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

/**
 *
 * @param bh_context
 * @return
 */

sync_envelope_msg_type GetState(const void *bh_context) {
	sync_envelope_err_type status = SYNC_ERR_FAIL;

	if (bh_context == NULL)
		return status;

	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	status = sync_envelope_get_state(bh_context);

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

/**
 *
 * @param bh_context
 * @param new_state
 * @return
 */
sync_envelope_err_type SetState(const void *bh_context, sync_envelope_msg_type new_state)
{
	sync_envelope_err_type status = SYNC_ERR_FAIL;

	if (bh_context == NULL) {
		return status;
	} else {
		/*Acquired Binary Semaphore*/
		xSemaphoreTake(g_bh_binary_semaphore0, 0);

		status = sync_envelope_set_state(bh_context, new_state);

		/*Release Binary Semaphore*/
		xSemaphoreGive(g_bh_binary_semaphore0);
	}

	return status;

}

/**
 *
 * @param bh_context
 * @param payload
 * @param payload_len
 * @param outBuffer
 * @param oBufferLen
 * @return
 */

sync_envelope_err_type DeviceHeartBeat(void *bh_context, uint8_t *payload, uint32_t payload_len,
		 uint8_t *outBuffer, uint32_t *oBufferLen)
{
	sync_envelope_err_type status = SYNC_ERR_FAIL;
	uint32_t buf_len = 0;
	uint8_t frame_type = SYNC_ENVELOPE_SIGNED;
	uint8_t msg_type = DEVICE_HEARTBEAT;

	if (bh_context == NULL || payload == NULL)
		return SYNC_ERR_INVALID_HANDLE;

	/*Acquired Binary Semaphore*/
	xSemaphoreTake(g_bh_binary_semaphore0, 0);

	/*create discovery block*/
	buf_len = (uint32_t) sync_envelope_process_payload(bh_context, frame_type,
			msg_type, payload, payload_len, outBuffer);

	if (buf_len > 0) {
		status = SYNC_OK;
		*oBufferLen = buf_len;
	}

	/*Release Binary Semaphore*/
	xSemaphoreGive(g_bh_binary_semaphore0);

	return status;
}

